import unittest

from metapub import PubMedFetcher
from metapub.pubmedcentral import *

class TestPubmedFetcher(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_article_by_pmid(self):
        pmid = '4'
        fetch = PubMedFetcher()
        article = fetch.article_by_pmid(pmid)
        assert str(article.pmid) == pmid

        pmid = '25763451'
        fetch = PubMedFetcher()
        article = fetch.article_by_pmid(pmid)
        assert str(article.pmid) == pmid



from __future__ import absolute_import, print_function

__doc__ = '''metapub.PubMedFetcher -- tools to deal with NCBI's E-utilities interface to PubMed'''
__author__ = 'nthmost'

from eutils.exceptions import EutilsBadRequestError
from lxml import etree
import requests

from .pubmedarticle import PubMedArticle
from .pubmedcentral import get_pmid_for_otherid
from .utils import kpick, parameterize, lowercase_keys
from .text_mining import re_pmid
from .exceptions import *
from .base import Borg
from .config import DEFAULT_EMAIL

def get_uids_from_esearch_result(xmlstr):
    dom = etree.fromstring(xmlstr)
    uids = []
    idlist = dom.find('IdList')
    for item in idlist.findall('Id'):
        uids.append(item.text.strip())
    return uids

class PubMedFetcher(Borg):
    '''PubMedFetcher (a Borg singleton object)

    An interaction layer for querying via specified method to return PubMedArticle objects.
    
    Currently available methods: eutils

    Basic Usage:

        fetch = PubMedFetcher()
    
    To specify a service method (more coming soon):

        fetch = PubMedFetcher('eutils')

    To return an article by querying the service with a known PMID:

        paper = fetch.article_by_pmid('123456')

    Similar methods exist for returning papers by DOI and PM Central id:

        paper = fetch.article_by_doi('10.1038/ng.379')
        paper = fetch.article_by_pmcid('PMC3458974')

    Finally, you can search for PMIDs via citation details by using the pmids_for_citation
    method, for which you only need 3 out of 5 details as a list of one or more PMIDs:

        pmids = fetch.pmids_for_citation(journal='Science', year='2008', volume='4', 
                first_page='7', author_name='Grant')
    '''

    def __init__(self, method='eutils', email=DEFAULT_EMAIL):
        Borg.__init__(self)
        self.method = method

        if method=='eutils':
            import eutils.client as ec
            self.qs = ec.QueryService(tool='metapub', email=email)
            self.article_by_pmid = self._eutils_article_by_pmid
            self.article_by_pmcid = self._eutils_article_by_pmcid
            self.article_by_doi = self._eutils_article_by_doi
            self.pmids_for_query = self._eutils_pmids_for_query
        else:
            raise NotImplementedError('coming soon: fetch from local pubmed via medgen-mysql or filesystem cache.')

    def _eutils_article_by_pmid(self, pmid):
        pmid = str(pmid)
        try:
            result = self.qs.efetch(args={'db': 'pubmed', 'id': pmid})
        except EutilsBadRequestError:
            raise MetaPubError('Invalid ID "%s" (rejected by Eutils); please check the number and try again.' % pmid)

        if result==None:
            return None
        if result.find('ERROR') > -1:
            raise MetaPubError('PMID %s returned ERROR; cannot construct PubMedArticle' % pmid)

        try:
            return PubMedArticle(result)
        except AttributeError:
            # in this case, eutils let us fetch a good-looking pmid, but it did not parse as an article.
            raise InvalidPMID('Pubmed ID "%s" not found' % pmid)

    def _eutils_article_by_pmcid(self, pmcid):
        # if user submitted a bare number, prepend "PMC" to make sure it is submitted correctly 
        # the conversion API at pubmedcentral.
        pmcid = str(pmcid)
        if re_pmid.findall(pmcid)[0] == pmcid:
            pmcid = 'PMC'+pmcid

        pmid = get_pmid_for_otherid(pmcid)
        if pmid is None:
            raise MetaPubError('No PMID available for PubMedCentral id %s' % pmcid)
        return self._eutils_article_by_pmid(pmid)
    
    def _eutils_article_by_doi(self, doi):
        pmid = get_pmid_for_otherid(doi)
        if pmid is None:
            raise MetaPubError('No PMID available for doi %s' % doi)
        return self._eutils_article_by_pmid(pmid)

    def _eutils_pmids_for_query(self, query='', since=None, until=None, retstart=0, retmax=250,
                pmc_only=False, **kwargs):
            
        '''returns list of pmids for given freeform query string plus keyword arguments.
            
        All Pubmed Advanced Query tokens (e.g. "TI" for title) are supported, plus many 
        "soft" keywords that will, when parsed, map to the correct token for you. For example,
        you can supply journal name with any of the following keywords:

            "journal" == "jtitle" == "journal_title" == "TA"

        Keyword arguments are case-INsensitive since they are lowercased upon arrival.

        Example of series of queries to accomplish "pagination" of data:

        first_250 = fetch.pmids_for_query('some query')
        second_250 = fetch.pmids_for_query('some query', retstart=500, retmax=250)

        :param: query (string) default ''
        :param: since (string) default None  # Y/m/d format expected. Y alone or Y/m allowed.
        :param: until (string) default None  # Y/m/d format expected. Y alone or Y/m allowed.
        :param: retstart (int) default 0
        :param: retmax (int) default 250
        :param: pmc_only (bool) default False  # constructs query to only search Pubmed Central.
        
        '''

        # lowercase all the things.
        kwargs = lowercase_keys(kwargs)

        q = {}
        if query:
            q['ALL'] = query

        # Search within date range (since / until)
        # 
        # working examples. search by creation date only works within defined ranges (not "< X" or "> Y")
        # ("2015/3/1"[Date - Create] : "2015/3/3"[Date - Create]) 
        # ("2015/2/14"[CRDT] : "2015/3/14"[CRDT])
        created_date_template = '"%s"[CRDT]'
        date_range_template = " (%s : %s)"
        if since:
            start = created_date_template % since
            if until:
                end = created_date_template % until
            else:
                end = '"3000"[CRDT]'

            query += date_range_template % (start, end)

        # unique ID referents.
        q['PMID'] = kpick(kwargs, options=['pmid', 'uid', 'pubmed_id'])
        q['AID'] = kpick(kwargs, options=['aid', 'doi']) 
        q['book'] = kwargs.get('book', None)
        q['JID'] = kpick(kwargs, options=['jid', 'nlm uid', 'nlm unique id'])
        q['ISBN'] = kwargs.get('ISBN', None)
        q['RN'] = kpick(kwargs, options=['rn', 'rcn', 'ecn'])
        q['GR'] = kpick(kwargs, options=['gr', 'grant number'])

        # Pubmed Date features:
        q['DA'] = kpick(kwargs, options=['da', 'date created'])
        q['LR'] = kpick(kwargs, options=['lr', 'date revised', 'date last revised'])
        q['EDAT'] = kpick(kwargs, options=['edat', 'entrez date'])

        # Journal name:
        q['TA'] = kpick(kwargs, options=['ta', 'journal', 'jtitle', 'journal_title'])

        # Article-level characteristics (title, authors, etc):
        q['TIAB'] = kpick(kwargs, options=['tiab', 'abstract', 'title/abstract'])
        q['TI'] = kpick(kwargs, options=['ti', 'title', 'atitle', 'article_title'])
        q['TT'] = kpick(kwargs, options=['tt', 'transliterated title'])

        q['AU'] = kpick(kwargs, options=['au', 'author'])
        q['1AU'] = kpick(kwargs, options=['1au', 'aulast', 'author1_lastfm', 'author1_last_fm'])
        q['FAU'] = kpick(kwargs, options=['fau', 'first_author', 'author1'])
        q['LASTAU'] = kpick(kwargs, options=['lastau', 'last author'])
        q['CN'] = kpick(kwargs, options=['cn', 'corporate author'])
        q['FIR'] = kpick(kwargs, options=['fir', 'full investigator name'])
        q['IR'] = kpick(kwargs, options=['ir', 'investigator'])
        q['PG'] = kpick(kwargs, options=['pg', 'pages', 'spage', 'first_page'])

        # Volume / Issue characteristics
        q['IP'] = kpick(kwargs, options=['ip', 'issue'])
        q['VTI'] = kpick(kwargs, options=['vta', 'volume title'])
        q['VI'] = kpick(kwargs, options=['vi', 'volume', 'vol'])

        # Content characteristics
        q['LA'] = kpick(kwargs, options=['la', 'language'])
        q['TW'] = kpick(kwargs, options=['tw', 'text'])
        q['PS'] = kpick(kwargs, options=['ps', 'personal name as subject']) 
        q['PA'] = kpick(kwargs, options=['pa', 'pharmacological action'])
        q['SB'] = kpick(kwargs, options=['sb', 'subset'])
        q['NM'] = kpick(kwargs, options=['nm', 'supplementary concept'])

        # MeSH characteristics 
        q['MHDA'] = kpick(kwargs, options=['mhda', 'mesh date'])
        q['MH'] = kpick(kwargs, options=['mh', 'mesh', 'mesh terms'])
        q['MAJR'] = kpick(kwargs, options=['majr', 'mesh major topic', 'mesh major'])
        q['SH'] = kpick(kwargs, options=['sh', 'mesh subheadings'])

        # Publication characteristics
        q['DCOM'] = kpick(kwargs, options=['dcom', 'completion date'])
        q['DP'] = kpick(kwargs, options=['dp', 'date of publication', 'year', 'pdat']) #most aligned w/ PubMedArticle.year and CrossRef 'year'
        q['LID'] = kpick(kwargs, options=['lid', 'location id', 'location identifier'])
        q['PUBN'] = kpick(kwargs, options=['pubn', 'publisher'])
        q['PT'] = kpick(kwargs, options=['pt', 'pubmed_type', 'publication type'])
        q['PL'] = kpick(kwargs, options=['pk', 'place of publication'])

        # Miscellaneous, alphabetized by Medline feature tag.
        q['AD'] = kpick(kwargs, options=['ad', 'affiliation']) 
        q['OT'] = kpick(kwargs, options=['ot', 'other term'])
        q['NM'] = kpick(kwargs, options=['nm', 'substance Name'])
        q['SI'] = kpick(kwargs, options=['si', 'secondary source id']) 

        for feature in q.keys():
            if q[feature] != None:
                query +=' "%s"[%s]' % (q[feature], feature)
        
        # option to query pubmed central only:
        # pubmed pmc[sb]
        if pmc_only:
            query += ' pubmed pmc[sb]'

        result = self.qs.esearch({ 'db': 'pubmed', 'term': query, 
                                   'retmax': retmax, 'retstart': retstart })
        return get_uids_from_esearch_result(result)

    def pmids_for_citation(self, **kwargs):
        '''returns list of pmids for given citation. requires at least 3/5 of these keyword arguments:
            jtitle or journal (journal title)
            year or date
            volume
            spage or first_page (starting page / first page)
            aulast (first author's last name) or author1_first_lastfm (as produced by PubMedArticle class)

        (Note that these arguments were made to match the tokens that arise from CrossRef's result['slugs'].)
        '''
        # output format in return:
        # journal_title|year|volume|first_page|author_name|your_key|
        base_uri = 'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/ecitmatch.cgi?db=pubmed&retmode=xml&bdata={journal_title}|{year}|{volume}|{first_page}|{author_name}|metapub|'

        kwargs = lowercase_keys(kwargs)
        journal_title = kpick(kwargs, options=['jtitle', 'journal', 'journal_title'], default='')
        author_name = _reduce_author_string(kpick(kwargs, 
                        options=['aulast', 'author1_last_fm', 'author', 'authors'], default=''))
        first_page = kpick(kwargs, options=['spage', 'first_page'], default='')
        year = kpick(kwargs, options=['year', 'date', 'pdat'], default='')
        volume = kpick(kwargs, options=['volume'], default='')

        inp_dict = { 'journal_title': parameterize(journal_title, '+'), 
                     'year': str(year),
                     'volume': str(volume),
                     'first_page': str(first_page),
                     'author_name': parameterize(author_name, '+'),
                   }

        req = base_uri.format(**inp_dict)
        content = requests.get(req).text
        pmids = []
        for item in content.split('\n'):
            if item.strip():
                pmid = item.split('|')[-1]
                pmids.append(pmid.strip())
        return pmids
                

def _reduce_author_string(author_string):
    # try splitting by commas
    authors = author_string.split(',')
    if len(authors)<2:
        # try splitting by semicolons
        authors = author_string.split(';')

    author1 = authors[0]
    # presume last name is at the end of the string
    return author1.split(' ')[-1]


""" 
Search Field Descriptions and Tags

from http://www.ncbi.nlm.nih.gov/books/NBK3827/

Affiliation [AD]
Article Identifier [AID]
All Fields [ALL]
Author [AU]
Author Identifier [AUID]
Book [book]
Comment Corrections
Corporate Author [CN]
Create Date [CRDT]
Completion Date [DCOM]
EC/RN Number [RN]
Editor [ED]
Entrez Date [EDAT]
Filter [FILTER]
First Author Name [1AU]
Full Author Name [FAU]
Full Investigator Name [FIR]
Grant Number [GR]   Investigator [IR]
ISBN [ISBN]
Issue [IP]
Journal [TA]
Language [LA]
Last Author [LASTAU]
Location ID [LID]
MeSH Date [MHDA]
MeSH Major Topic [MAJR]
MeSH Subheadings [SH]
MeSH Terms [MH]
Modification Date [LR]
NLM Unique ID [JID]
Other Term [OT]
Owner
Pagination [PG]
Personal Name as Subject [PS]   Pharmacological Action [PA]
Place of Publication [PL]
PMID [PMID]
Publisher [PUBN]
Publication Date [DP]
Publication Type [PT]
Secondary Source ID [SI]
Subset [SB]
Supplementary Concept[NM]
Text Words [TW]
Title [TI]
Title/Abstract [TIAB]
Transliterated Title [TT]
UID [PMID]
Version
Volume [VI]
"""

'''<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE eSearchResult PUBLIC "-//NLM//DTD esearch 20060628//EN" "http://eutils.ncbi.nlm.nih.gov/eutils/dtd/20060628/esearch.dtd">
<eSearchResult><Count>1</Count><RetMax>1</RetMax><RetStart>0</RetStart><QueryKey>1</QueryKey><WebEnv>NCID_1_129952677_165.112.9.37_9001_1426564126_1266564592_0MetA0_S_MegaStore_F_1</WebEnv><IdList>
<Id>25023161</Id>
</IdList><TranslationSet><Translation>     <From>Journal of Neural Transmission[TA]</From>     <To>"J Neural Transm"[Journal] OR "J Neural Transm"[Journal] OR "J Neural Transm Suppl"[Journal] OR "J Neural Transm Park Dis Dement Sect"[Journal] OR "J Neural Transm Gen Sect"[Journal]</To>    </Translation></TranslationSet><TranslationStack>   <TermSet>    <Term>2014[DP]</Term>    <Field>DP</Field>    <Count>1171381</Count>    <Explode>N</Explode>   </TermSet>   <TermSet>    <Term>"J Neural Transm"[Journal]</Term>    <Field>Journal</Field>    <Count>1177</Count>    <Explode>N</Explode>   </TermSet>   <TermSet>    <Term>"J Neural Transm"[Journal]</Term>    <Field>Journal</Field>    <Count>3005</Count>    <Explode>N</Explode>   </TermSet>   <OP>OR</OP>   <TermSet>    <Term>"J Neural Transm Suppl"[Journal]</Term>    <Field>Journal</Field>    <Count>1409</Count>    <Explode>N</Explode>   </TermSet>   <OP>OR</OP>   <TermSet>    <Term>"J Neural Transm Park Dis Dement Sect"[Journal]</Term>    <Field>Journal</Field>    <Count>226</Count>    <Explode>N</Explode>   </TermSet>   <OP>OR</OP>   <TermSet>    <Term>"J Neural Transm Gen Sect"[Journal]</Term>    <Field>Journal</Field>    <Count>526</Count>    <Explode>N</Explode>   </TermSet>   <OP>OR</OP>   <OP>GROUP</OP>   <OP>AND</OP>   <TermSet>    <Term>121[VI]</Term>    <Field>VI</Field>    <Count>44530</Count>    <Explode>N</Explode>   </TermSet>   <OP>AND</OP>   <TermSet>    <Term>Freitag[1AU]</Term>    <Field>1AU</Field>    <Count>496</Count>    <Explode>N</Explode>   </TermSet>   <OP>AND</OP>  </TranslationStack><QueryTranslation>2014[DP] AND ("J Neural Transm"[Journal] OR "J Neural Transm"[Journal] OR "J Neural Transm Suppl"[Journal] OR "J Neural Transm Park Dis Dement Sect"[Journal] OR "J Neural Transm Gen Sect"[Journal]) AND 121[VI] AND Freitag[1AU]</QueryTranslation></eSearchResult>'''



from __future__ import absolute_import

"""metapub.medgenconcept -- MedGenConcept class instantiated by supplying ESummary XML string."""

import logging
from lxml import etree

from .base import MetaPubObject
from .exceptions import MetaPubError

logger = logging.getLogger()

class MedGenConcept(MetaPubObject):

    def __init__(self, xmlstr, *args, **kwargs):
        super(MedGenConcept, self).__init__(xmlstr, 'DocumentSummarySet/DocumentSummary', args, kwargs)

        if self._get('error'):
            raise MetaPubError('Supplied XML for MedGenConcept contained explicit error: %s' % self._get('error') )

        # sometimes, ConceptMeta is an XML document embedded within the XML response. Boo-urns. 
        try:
            self.meta = etree.fromstring('<ConceptMeta>'+self.content.find('ConceptMeta').text+'</ConceptMeta>')
        except TypeError:
            self.meta = self.content.find('ConceptMeta')
        
        self.modes_of_inheritance = self._get_modes_of_inheritance()
        self.OMIM = self._get_OMIM()        # is a list, since sometimes there are more than one.
        self.names = self._get_names()
        self.CUI = self._get_CUI()
        self.title = self._get_title()
        self.definition = self._get_definition()
        self.semantic_id = self._get_semantic_id()
        self.semantic_type = self._get_semantic_type()
        self.associated_genes = self._get_associated_genes()
        self.cytogenic = self._get_cytogenic()
        self.chromosome = self._get_chromosome()
        self.medgen_uid = self._get_medgen_uid()


    def to_dict(self):
        'returns a dictionary composed of all extractable properties of this concept.'
        return { 'CUI': self.CUI, 'title': self.title, 'definition': self.definition,
                 'semantic_id': self.semantic_id, 'semantic_type': self.semantic_type,
                 'modes_of_inheritance': self.modes_of_inheritance, 
                 'associated_genes': self.associated_genes, 'medgen_uid': self.medgen_uid,
                 'names': self.names, 'OMIM': self.OMIM, 'cytogenic': self.cytogenic,
                 'chromosome': self.chromosome }

    def _get_CUI(self):
        return self._get('ConceptId')
    
    def _get_title(self):
        return self._get('Title')
        
    def _get_definition(self):
        return self._get('Definition')
    
    def _get_semantic_id(self):
        return self._get('SemanticId')
    
    def _get_semantic_type(self):
        return self._get('SemanticType')
        
    def _get_medgen_uid(self):
        return self.content.get('uid')
    
    def _get_modes_of_inheritance(self):
    
        '''returns a list of all known ModesOfInheritance, in format:
        [ { 'CUI': 'CNxxxx', 'name': 'some name', 'medgen_uid': 'xxxxxx', 'tui': 'A000 }, ...  ]
        '''
        output_list = []
        modes = self.meta.find('ModesOfInheritance').getchildren()
        
        extra_key_dict = { 'CUI': None, 
                           'TUI': None,
                           'medgen_uid': None, 
                         }
        for mode in modes:
            mode_dict = extra_key_dict.copy()
            try:
                mode_dict['semantic_type'] = mode.find('SemanticType').text
            except AttributeError:
                pass
            try:
                mode_dict['definition'] = mode.find('Definition').text
            except AttributeError:
                pass
            mode_dict['name'] = mode.find('Name').text

            for item in extra_key_dict.keys():
                try:
                    mode.get(item)
                except AttributeError:
                    pass
                
            output_list.append(mode_dict)
        return output_list
             
    def _get_associated_genes(self):
        '''returns a list of AssociatedGenes, in format:
        [ { 'gene_id': 'xxx', 'chromosome': 'X', 'cytogen_loc': 'X9234235', 'hgnc': 'GENE' }, ]
        
        if not available, returns None. 
        '''
        genes = []
        try:
            for gene in self.meta.find('AssociatedGenes').getchildren():
                genes.append({ 'gene_id': gene.get('gene_id'), 
                               'hgnc': gene.text,
                               'chromosome': gene.get('chromosome'),
                               'cytogen_loc': gene.get('cytogen_loc') })
            return genes
        except AttributeError:
            return None

    def _get_names(self):
        '''returns a list of this concept's equivalent Names in various dictionaries,
        in format:
        
        { 'SDUI': '300555', 'SCUI': 'xxx', 'CODE': '300555', 'SAB': 'OMIM' 'TTY': 'PT' 'type': 'syn', 'name': 'DENT DISEASE 2' }
        
        '''
        names = []

        # not every ID is present in each Name (e.g. SCUI only appears sometimes).        
        possible_keys = ['SDUI', 'SCUI', 'CODE', 'SAB', 'TTY', 'PT', 'type']
        
        for name in self.meta.find('Names').getchildren():
            outd = { 'name': name.text }
            for key in possible_keys:
                try:
                    outd[key] = name.get(key)            
                except AttributeError:
                    pass
            names.append(outd)
        return names
        
    def _get_OMIM(self):
        '''returns this concept's OMIM ids (list of strings), when available, else returns [].'''
        omim_root = self.meta.find('OMIM')
        #from IPython import embed; embed()
        #print omim_root.find('MIM').text
        outp = []
        try:
            mims = omim_root.get_children()
        except AttributeError:
            return None

        for item in omim_root.get_children():
            try:
                outp.append(item.text)
            except AttributeError:
                pass
        return outp
        
    def _get_chromosome(self):
        '''returns this concept's affected chromosome, if applicable/available'''
        try:
            return self.meta.find('Chromosome').text
        except AttributeError:
            return None

    def _get_cytogenic(self):
        '''returns this concept's cytogenic property, if applicable/available'''
        try:
            return self.meta.find('Cytogenic').text
        except AttributeError:
            return None

    # TODO
    # Definitions
    # <Definitions><Definition source="GeneReviews">Dent disease, an X-linked disorder of proximal renal tubular dysfunction, is characterized by low molecular-weight (LMW) proteinuria, hypercalciuria, nephrocalcinosis, nephrolithiasis, and chronic kidney disease (CKD). Males younger than age ten years may manifest only low molecular-weight (LMW) proteinuria and/or hypercalciuria, which are usually asymptomatic. Thirty to 80% of affected males develop end-stage renal disease (ESRD) between ages 30 and 50 years; in some instances ESRD does not develop until the sixth decade of life or later. Rickets or osteomalacia are occasionally observed, and mild short stature, although underappreciated, may be a common occurrence. Disease severity can vary within the same family. Males with Dent disease 2 (caused by mutation of OCRL) are at increased risk for intellectual disability. Due to random X-chromosome inactivation, some female carriers may manifest hypercalciuria and, rarely, renal calculi and moderate LMW proteinuria. Females rarely if ever develop CKD.</Definition></Definitions>
    
    # TODO
    # ClinicalFeatures / ClinicalFeature
    # <ClinicalFeatures><ClinicalFeature uid="9232" CUI="C0019322" TUI="T190" SDUI="HP:0001537"><Name>Umbilical hernia</Name><SemanticType>Anatomical Abnormality</SemanticType></ClinicalFeature><ClinicalFeature uid="87607" CUI="C0349588" TUI="T033" SDUI="HP:0004322"><Name>Short stature</Name><SemanticType>Finding</SemanticType></ClinicalFeature><ClinicalFeature uid="333360" CUI="C1839606" TUI="T033" SDUI="HP:0003126"><Name>Low-molecular-weight proteinuria</Name><SemanticType>Finding</SemanticType></ClinicalFeature><ClinicalFeature uid="383844" CUI="C1856145" TUI="T033" SDUI="HP:0100543"><Name>Cognitive impairment</Name><SemanticType>Finding</SemanticType></ClinicalFeature><ClinicalFeature uid="349145" CUI="C1859342" TUI="T033" SDUI="HP:0000114"><Name>Proximal tubulopathy</Name><SemanticType>Finding</SemanticType></ClinicalFeature><ClinicalFeature uid="504348" CUI="CN000117" TUI="T033" SDUI="HP:0000121"><Name>Nephrocalcinosis</Name><SemanticType>Finding</SemanticType><Definition>Nephrocalcinosis is the deposition of calcium salts in renal parenchyma.</Definition></ClinicalFeature><ClinicalFeature uid="504774" CUI="CN001157" TUI="T033" SDUI="HP:0001263"><Name>Global developmental delay</Name><SemanticType>Finding</SemanticType><Definition>A delay in the achievement of motor or mental milestones in the domains of development of a child, including motor skills, speech and language, cognitive skills, and social and emotional skills. This term should only be used to describe children younger than five years of age.</Definition></ClinicalFeature><ClinicalFeature uid="505127" CUI="CN001948" TUI="T033" SDUI="HP:0002150"><Name>Hypercalciuria</Name><SemanticType>Finding</SemanticType></ClinicalFeature><ClinicalFeature uid="505493" CUI="CN002923" TUI="T033" SDUI="HP:0003236"><Name>Elevated serum creatine phosphokinase</Name><SemanticType>Finding</SemanticType><Definition>An elevation of the level of the enzyme creatine kinase (also known as creatine phosphokinase, CPK; EC 2.7.3.2) in the blood. CPK levels can be elevated in a number of clinical disorders such as myocardial infarction, rhabdomyolysis, and muscular dystrophy.</Definition></ClinicalFeature><ClinicalFeature uid="425142" CUI="CN003029" TUI="T033" SDUI="HP:0003355"><Name>Aminoaciduria</Name><SemanticType>Finding</SemanticType><Definition>An increased concentration of an amino acid in the urine.</Definition></ClinicalFeature><ClinicalFeature uid="776439" CUI="CN183891" TUI="T033" SDUI="HP:0012622"><Name>Chronic kidney disease</Name><SemanticType>Finding</SemanticType><Definition>Functional anomaly of the kidney persisting for at least three months.</Definition></ClinicalFeature></ClinicalFeatures><PhenotypicAbnormalities><Category CUI="CN000115" name="Abnormality of the genitourinary system"><ClinicalFeature uid="504348" CUI="CN000117" TUI="T033" SDUI="HP:0000121"><SemanticType>Finding</SemanticType><Name>Nephrocalcinosis</Name><Definition>Nephrocalcinosis is the deposition of calcium salts in renal parenchyma.</Definition></ClinicalFeature><ClinicalFeature uid="425142" CUI="CN003029" TUI="T033" SDUI="HP:0003355"><SemanticType>Finding</SemanticType><Name>Aminoaciduria</Name><Definition>An increased concentration of an amino acid in the urine.</Definition></ClinicalFeature><ClinicalFeature uid="776439" CUI="CN183891" TUI="T033" SDUI="HP:0012622"><SemanticType>Finding</SemanticType><Name>Chronic kidney disease</Name><Definition>Functional anomaly of the kidney persisting for at least three months.</Definition></ClinicalFeature></Category><Category CUI="CN000664" name="Abnormality of the nervous system"><ClinicalFeature uid="504774" CUI="CN001157" TUI="T033" SDUI="HP:0001263"><SemanticType>Finding</SemanticType><Name>Global developmental delay</Name><Definition>A delay in the achievement of motor or mental milestones in the domains of development of a child, including motor skills, speech and language, cognitive skills, and social and emotional skills. This term should only be used to describe children younger than five years of age.</Definition></ClinicalFeature></Category><Category CUI="CN001754" name="Abnormality of metabolism/homeostasis"><ClinicalFeature uid="505493" CUI="CN002923" TUI="T033" SDUI="HP:0003236"><SemanticType>Finding</SemanticType><Name>Elevated serum creatine phosphokinase</Name><Definition>An elevation of the level of the enzyme creatine kinase (also known as creatine phosphokinase, CPK; EC 2.7.3.2) in the blood. CPK levels can be elevated in a number of clinical disorders such as myocardial infarction, rhabdomyolysis, and muscular dystrophy.</Definition></ClinicalFeature><ClinicalFeature uid="425142" CUI="CN003029" TUI="T033" SDUI="HP:0003355"><SemanticType>Finding</SemanticType><Name>Aminoaciduria</Name><Definition>An increased concentration of an amino acid in the urine.</Definition></ClinicalFeature>
    
    # TODO
    # PhenotypicAbnormalities 
    # <PhenotypicAbnormalities><Category CUI="CN000115" name="Abnormality of the genitourinary system"><ClinicalFeature uid="504348" CUI="CN000117" TUI="T033" SDUI="HP:0000121"><SemanticType>Finding</SemanticType><Name>Nephrocalcinosis</Name><Definition>Nephrocalcinosis is the deposition of calcium salts in renal parenchyma.</Definition></ClinicalFeature><ClinicalFeature uid="425142" CUI="CN003029" TUI="T033" SDUI="HP:0003355"><SemanticType>Finding</SemanticType><Name>Aminoaciduria</Name><Definition>An increased concentration of an amino acid in the urine.</Definition></ClinicalFeature><ClinicalFeature uid="776439" CUI="CN183891" TUI="T033" SDUI="HP:0012622"><SemanticType>Finding</SemanticType><Name>Chronic kidney disease</Name><Definition>Functional anomaly of the kidney persisting for at least three months.</Definition></ClinicalFeature></Category><Category CUI="CN000664" name="Abnormality of the nervous system"><ClinicalFeature uid="504774" CUI="CN001157" TUI="T033" SDUI="HP:0001263"><SemanticType>Finding</SemanticType><Name>Global developmental delay</Name><Definition>A delay in the achievement of motor or mental milestones in the domains of development of a child, including motor skills, speech and language, cognitive skills, and social and emotional skills. This term should only be used to describe children younger than five years of age.</Definition></ClinicalFeature></Category><Category CUI="CN001754" name="Abnormality of metabolism/homeostasis"><ClinicalFeature uid="505493" CUI="CN002923" TUI="T033" SDUI="HP:0003236"><SemanticType>Finding</SemanticType><Name>Elevated serum creatine phosphokinase</Name><Definition>An elevation of the level of the enzyme creatine kinase (also known as creatine phosphokinase, CPK; EC 2.7.3.2) in the blood. CPK levels can be elevated in a number of clinical disorders such as myocardial infarction, rhabdomyolysis, and muscular dystrophy.</Definition></ClinicalFeature><ClinicalFeature uid="425142" CUI="CN003029" TUI="T033" SDUI="HP:0003355"><SemanticType>Finding</SemanticType><Name>Aminoaciduria</Name><Definition>An increased concentration of an amino acid in the urine.</Definition></ClinicalFeature></Category></PhenotypicAbnormalities>
        
    # TODO
    # <RelatedDisorders></RelatedDisorders>
    
    # TODO
    # <SNOMEDCT></SNOMEDCT>
    
    # known others not planned for inclusion:
    # <PharmacologicResponse></PharmacologicResponse>
    



import unittest

from metapub import MedGenFetcher
fetch = MedGenFetcher()

hugos = ['ACVRL1']

class TestPubmedSearch(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_fetch_concepts_for_known_gene(self):
        hugo = 'ACVRL1'
        result = fetch.uids_by_term(hugo+'[gene]')
        assert result is not None
        assert result[0]=='324960'
    
    def test_fetch_concepts_for_incorrect_term(self):
        term = 'AVCRL'
        result = fetch.uids_by_term(term+'[gene]')
        assert result==[]




from __future__ import absolute_import

import os
import logging

PKGNAME='metapub'

# where to place XML (temporarily) when downloaded.
TMPDIR = '/tmp'

# email address submitted to eutils with requests (as required by their api).
DEFAULT_EMAIL=os.getenv('EUTILS_EMAIL', 'naomi.most@invitae.com')

#### keep eutils yammering down to a reasonable level.
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("eutils").setLevel(logging.INFO)
####

def get_process_log(filepath, loglevel=logging.INFO, name=PKGNAME+'-process'):
    log = logging.getLogger(name)
    log.setLevel(loglevel)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    fh = logging.FileHandler(filepath)
    fh.setFormatter(formatter)
    fh.setLevel(loglevel)
    log.addHandler(fh)
    return log

def get_data_log(filepath, name=PKGNAME+'-data'):
    datalog = logging.getLogger(name)
    datalog.setLevel(logging.DEBUG)
    datalog.propagate = False
    formatter = logging.Formatter('')
    fh = logging.FileHandler(filepath)
    datalog.addHandler(fh)
    return datalog




# Biomed/api/stuff/gene2disease/$GENE
#
# expected response includes 
#
# SemanticType=='Disease or Syndrome' 
# MedGenUID=='C####' | 'CN####' | #### | Example: 2881
# 
# Subtypes: 
# For this disease, what are the subtypes? 
# Subtype::= MedGenUID 'Disease or Syndrome' where MGREL has_parent==2881  
#
#
# Hypertrophic Cardiomyopathy
# http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi?db=medgen&id=2881
#
# Gene OCRL
# http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi?db=medgen&term=OCRL
# result:

"""<eSearchResult>
<Count>5</Count>
<RetMax>5</RetMax>
<RetStart>0</RetStart>
<IdList>
<Id>763754</Id>
<Id>336867</Id>
<Id>336322</Id>
<Id>168056</Id>
<Id>18145</Id>
</IdList>
<TranslationSet/>
<TranslationStack>
<TermSet>
<Term>OCRL[All Fields]</Term>
<Field>All Fields</Field>
<Count>5</Count>
<Explode>N</Explode>
</TermSet>
<OP>GROUP</OP>
</TranslationStack>
<QueryTranslation>OCRL[All Fields]</QueryTranslation>
</eSearchResult>
"""

# from above list, get summary for id #18145
# http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi?db=medgen&id=18145
# result begins:
"""
<eSummaryResult>
<DocumentSummarySet status="OK">
<DbBuild>Build141124-1357.1</DbBuild>
<DocumentSummary uid="18145">
<ConceptId>C0028860</ConceptId>
<Title>Lowe syndrome</Title>
<Definition>
Lowe syndrome (oculocerebrorenal syndrome) is characterized by involvement of the eyes, central nervous system, and kidneys. Dense congenital cataracts are found in all affected boys and infantile glaucoma in approximately 50%. All boys have impaired vision; corrected acuity is rarely better than 20/100. Generalized hypotonia is noted at birth and is of central (brain) origin. Deep tendon reflexes are usually absent. Hypotonia may slowly improve with age, but normal motor tone and strength are never achieved. Motor milestones are delayed. Almost all affected males have some degree of intellectual disability; 10%-25% function in the low-normal or borderline range, approximately 25% in the mild-to-moderate range, and 50%-65% in the severe-to-profound range of intellectual disability. Affected males have varying degrees of proximal renal tubular dysfunction of the Fanconi type, including bicarbonate wasting and renal tubular acidosis, phosphaturia with hypophosphatemia and renal rickets, aminoaciduria, low molecular-weight (LMW) proteinuria, sodium and potassium wasting, and polyuria. Fanconi syndrome is usually not clinically apparent in the first few months of life, but symptoms may appear by age six to 12 months. Glomerulosclerosis associated with chronic tubular injury usually results in slowly progressive chronic renal failure and end-stage renal disease after age ten to 20 years.
</Definition>
<SemanticId>T047</SemanticId>
<SemanticType>Disease or Syndrome</SemanticType>
<Suppressed/>
<ConceptMeta>
...
STUFF
</ConceptMeta>
<ModificationDate/>
<Merged/>
</DocumentSummary>
</DocumentSummarySet>
</eSummaryResult>
"""


test_gene_list = """
FMR1
SYNE1
SYNE2
TMEM43
CRYAB
PCSK9
APOB
LDLRAP1
LDLR
APOE
F5
LIPC
FOXP3
PTPN22
"""




from metapub import PubMedFetcher
fetch = PubMedFetcher()
params = { 'jtitle': 'American Journal of Medical Genetics', 
                    'year': 1996, 
                    'volume': 61, 
                    'author1_lastfm': 'Hegmann' }

stuff = fetch.pmids_for_query(**params)
print params
print stuff

# the following article was deleted from pubmed (or changed such that this
# set of parameters no longer returns an article)
params = { 'TA':'Journal of Neural Transmission', 
                    'pdat':2014, 
                    'vol':121, 
                    'aulast': 'Freitag'
         } 

stuff = fetch.pmids_for_query(**params)

print params
print stuff

#params = { 'mesh': 'breast neoplasm' }
#stuff = fetch.pmids_for_query(since='2014', until='2015/3/1', pmc_only=True, **params)

pmids = fetch.pmids_for_query(since='2015/3/1', retmax=1000)
assert len(pmids)==1000




# -*- coding: utf-8 -*-

from __future__ import print_function

import os, sys, shutil
import logging, time

from requests.exceptions import ConnectionError

from metapub import PubMedFetcher, CrossRef 
from metapub.exceptions import XMLSyntaxError, MetaPubError
from metapub.text_mining import find_doi_in_string
from metapub.config import get_data_log
from metapub.pubmedcentral import get_pmcid_for_otherid
from metapub.utils import asciify

# configurables (edit as you like):
RESULTS_DIR = 'sandbox/results/'
DEBUG = False
CONNECTION_ERROR_LIMIT = 10

# non-configurables (don't touch):
error_log = None
data_log = None
CONNECTION_ERRORS = 0

# process_log: logs processing messages to stdout
process_log = logging.getLogger('pmid2doi')
ch = logging.StreamHandler()
if DEBUG:
    ch.setLevel(logging.DEBUG)
else:
    ch.setLevel(logging.INFO)
formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
ch.setFormatter(formatter)
process_log.addHandler(ch)
process_log.addHandler(ch)


def get_filename(base_fpath):
    def check_and_bump(x):
        if os.path.exists('%s.%i' % (base_fpath, x)):
            return check_and_bump(x+1)
        else:
            return '%s.%i' % (base_fpath, x)
    return check_and_bump(0)

def open_logs(run_span):
    global error_log
    global data_log
    fname_base = '%s.%s.log'
    errorlog_path = get_filename(os.path.join(RESULTS_DIR, fname_base % (run_span, 'error')))
    mapping_path = get_filename(os.path.join(RESULTS_DIR, fname_base % (run_span, 'mapping')))
    error_log = get_data_log(errorlog_path, 'pmid2doi.error')
    data_log = get_data_log(mapping_path, 'pmid2doi.mapping')
 

fetch = PubMedFetcher()
crossref = CrossRef()

def record_error(pmid, source, e):
    error_log.info('%s, %s, %r' % (pmid, source, e))

def record_entry(**kwargs):
    #{'doi': 'NA', 'score': 0, 'cr_results': [], 'connection_errors': 0, 'pma': None, 'cr_top_result': None, 'pmid': 1234567}
    fmt = '{pmid},{doi},{score},{pubmed_type}'   #,{crossref_query},{crossref_coins}'
    data_log.info(fmt.format(**kwargs))


def handle_connection_error(debug_msg=''):
    global CONNECTION_ERRORS
    CONNECTION_ERRORS += 1
    if CONNECTION_ERRORS > CONNECTION_ERROR_LIMIT:
        if debug_msg:
            process_log.debug(debug_msg)
        process_log.info('Too many connection errors (%i); quitting!' % CONNECTION_ERRORS)
        sys.exit()


class PMIDMapping(object):
    '''represents a mapping of a single pubmed ID (pmid) to as many IDs as 
        can be mapped to it.'''

    # number of times to retry getting XML from pubmed
    PMA_RETRY_LIMIT = 3

    def __init__(self, pmid):
        self.connection_errors = 0

        # set sensible defaults intended for display in records.
        self.pmid = pmid
        self.pma = None
        self.pm_type = None
        self.cr_top_result = None
        self.doi = 'NA'
        #self.pmcid = 'NA'       # not doing PMC just yet.
        self.score = 0          # becomes '10' if found in Pubmed, or a CrossRef score if found there.

        # 1) look up article in PubMed
        self._get_pma()         # fills self.pma

        if self.pma:
            # 2a) try to get doi first in PubMed, then in PMC, then in CrossRef.
            self._get_doi()     # fills self.doi, self.cr_top_result, self.score
            self.pm_type = self.pma.pubmed_type

            # 2b) try to get pmcid first in PubMed, then in PMC.
            # self._get_pmcid()
        else:
            self.score = -1

        # 3) record our efforts.
        outd = self.__dict__
        outd['crossref_query'] = asciify(crossref.last_query)
        outd['crossref_coins'] = None if not self.cr_top_result else self.cr_top_result['coins']
        outd['pubmed_type'] = self.pm_type

        record_entry(**outd)

    def _verify_doi(self, doi):
        return find_doi_in_string(doi)

    def _get_pma(self, tries=1):
        try:
            self.pma = fetch.article_by_pmid(self.pmid)
            process_log.debug("looking up %s (title: %s, journal: %s)" % (self.pmid, self.pma.title, self.pma.journal))

        except XMLSyntaxError:
            # XML did not completely download... try again.
            if tries < self.PMA_RETRY_LIMIT:
                tries += 1
                self._get_pma(tries)
        except ConnectionError:
            handle_connection_error()
        except Exception, e:
            record_error(self.pmid, 'PubMedArticle', e)

    def _get_pmcid(self):
        if self.pma.pmc:
            pmcid = self.pma.pmc
        else:
            pmcid = get_pmcid_for_otherid(pmid)
        if pmcid:
            self.pmcid = pmcid

    def _get_doi(self):
        if self.pma.doi:
            # there are quite a few DOIs are aren't real DOIs in PubMed XML.
            if self._verify_doi(self.pma.doi):
                process_log.debug("%s: Found DOI in PubMed XML", self.pma.pmid)
                self.doi = self.pma.doi
                self.score = 10
                self.source = 'PubMed'

        # No DOI yet?  keep going
        try:
            cr_results = crossref.query_from_PubMedArticle(self.pma)
        except ConnectionError:
            handle_connection_error('CrossRef could not be reached')
        except Exception, e:
            record_error(self.pmid, 'CrossRef', e)
            return

        if cr_results:
            self.top_result = crossref.get_top_result(cr_results)
        else:
            return        

        if self.top_result:
            self.doi = self.top_result['doi']
            self.score = float(self.top_result['score'])
            process_log.debug("%s: CrossRef found doi=%s with score=%.2f" % (self.pma, self.doi, self.score))
        else:
            process_log.debug("%s: CrossRef had no good results for query with q=%s" % (self.pmid, crossref.last_query))
            self.doi = 'NA'
            self.score = '0'


def main():
    pmid = stop = None
    try:
        pmid = int(sys.argv[1])
    except IndexError:
        print('Supply a STARTING PubMed ID number as the argument to this script.')
        sys.exit()
    except ValueError:
        print('First argument (PMID) must be an integer.')
        sys.exit()

    try:
        stop = int(sys.argv[2])
    except IndexError:
        print('No stopping point specified; default to +1000 from initial PMID.')
        stop = 1000
    except ValueError:
        print('Second argument (optional) must be an integer.')
        sys.exit()

    open_logs('%i_%i' % (pmid, stop))

    for pmid in xrange(pmid, pmid+stop):
        blah = PMIDMapping(pmid)
    

if __name__=='__main__':
    main()        





# -*- coding: utf-8 -*-

from __future__ import print_function

import os, sys, shutil
import logging
from urllib import unquote

from metapub import PubMedFetcher, CrossRef 
from metapub.exceptions import MetaPubError
from metapub.utils import asciify

#from tabulate import tabulate

####
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("eutils").setLevel(logging.INFO)
####

fetch = PubMedFetcher()

def print_table(res_table, headers):
    print('\t'.join(headers))
    for x in range(0, len(res_table['score'])):
        print('\t'.join(['{}'.format(res_table[h][x]) for h in headers]))

if __name__=='__main__':
    try:
        filename = sys.argv[1]
    except IndexError:
        print('Supply filename of pmid list as the argument to this script.')
        sys.exit()
    
    pmids = open(filename, 'r').readlines()

    results_table = { 'pmid': [], 'pma_title': [], 'cr_title': [], 'doi': [], 'score': [], 'pma_aulast': [], 'cr_aulast': [], 'pma_journal': [], 'cr_journal': [] } 

    CR = CrossRef()

    for pmid in pmids:
        pmid = pmid.strip()
        if pmid:
            results_table['pmid'].append(pmid)
            try:
                pma = fetch.article_by_pmid(pmid)
            except:
                pma = None
                #print("%s: Could not fetch" % pmid)
        if pma:
            results = CR.query_from_PubMedArticle(pma)
            top_result = CR.get_top_result(results, CR.last_params, use_best_guess=True)
                
            results_table['pma_title'].append(asciify(pma.title))
            results_table['pma_journal'].append(asciify(pma.journal))
            results_table['cr_title'].append(asciify(top_result['title']))
            results_table['doi'].append(top_result['doi'])
            results_table['score'].append(top_result['score'])

            if top_result['slugs'] != {}:
                try:
                    results_table['cr_aulast'].append(asciify(top_result['slugs']['aulast']))
                except:
                    results_table['cr_aulast'].append('NA')
                    
                results_table['cr_journal'].append(asciify(top_result['slugs']['jtitle']))
            else:
                results_table['cr_aulast'].append('')            
                results_table['cr_journal'].append('')

            #print(unquote(top_result['coins'])) #.decode('utf8'))

            results_table['pma_aulast'].append(asciify(pma.author1_last_fm))
        else:
            results = None
            top_result = None
            results_table['pma_aulast'].append('NA')
            results_table['pma_journal'].append('NA')
            results_table['pma_title'].append('NA')
            results_table['doi'].append('NA')
            results_table['score'].append('NA')
            results_table['cr_journal'].append('NA')
            results_table['cr_aulast'].append('NA')
            results_table['cr_title'].append('NA')

        #print(pmid, top_result['doi'], top_result['score'], sep='\t')

    headers = ['score', 'pmid', 'doi', 'pma_aulast', 'cr_aulast', 'pma_journal', 'cr_journal', 'pma_title', 'cr_title']

    print_table(results_table, headers)


# crossref return looks like:
"""{
        doi: "http://dx.doi.org/10.2307/40250596",
        score: 2.0651011,
        normalizedScore: 74,
        title: "Research and Relevant Knowledge: American Research Universities since World War II",
        fullCitation: "Winton U. Solberg, Roger L. Geiger, 1994, 'Research and Relevant Knowledge: American Research Universities since World War II', <i>Academe</i>, vol. 80, no. 1, p. 56",
        coins: "ctx_ver=Z39.88-2004&amp;rft_id=info%3Adoi%2Fhttp%3A%2F%2Fdx.doi.org%2F10.2307%2F40250596&amp;rfr_id=info%3Asid%2Fcrossref.org%3Asearch&amp;rft.atitle=Research+and+Relevant+Knowledge%3A+American+Research+Universities+since+World+War+II&amp;rft.jtitle=Academe&amp;rft.date=1994&amp;rft.volume=80&amp;rft.issue=1&amp;rft.spage=56&amp;rft.aufirst=Winton+U.&amp;rft.aulast=Solberg&amp;rft_val_fmt=info%3Aofi%2Ffmt%3Akev%3Amtx%3Ajournal&amp;rft.genre=article&amp;rft.au=Winton+U.+Solberg&amp;rft.au=+Roger+L.+Geiger",
        year: "1994"
        },"""

